/**
 * Egyszerű osztályhierarchia az ctor/dtor feladatainak bemutatásához
 * A pointerek átállítását csak közvetve, az f() virtuális függvényen
 * keresztül tudjuk megfigyelni, amit mindig a g() függvény hív az alaposztályból.
 * A g() függvényt viszont minden konstruktorból és destruktorból meghívunk.
 * Így Attól függően, hogy az f() virt. függvény pointere hova mutat,
 * más és más f() változat futnak le.
 */

 /**
 * HA a DUMP makró definiált, akkor a g() függvény az alaposztály
 * memóriaképét is kiírja hexadecimálisan.
 * FELTÉTELEZZÜK, hogy a memóriakép elején a virt. tábla mutatója áll.
 * Ezt követik az adattagok. Az egyszerűség kedvéért 3 db pointer az
 * adattag amit 1,2,3 értékkel töltünk fel.
 */
#define DUMP

#include <iostream>

using std::cout;
using std::cerr;
using std::cin;
using std::endl;

/**
 Konstruktor:
 * Öröklési lánc végén hívja a virtuális alaposztályok konstruktorait.
 * Hívja a közvetlen, nem virtuális alaposztályok konstruktorait.
 * Létrehozza a saját részt:
   - beállítja a virtuális alaposztály mutatóit
   - beállítja a virtuális függvények mutatóit
   - hívja a tartalmazott objektumok konstruktorait
   - végrehajtja a programozott törzset
 */


/**
 Destruktor:
 * Megszünteti a saját részt:
   - végrehajtja a programozott törzset
   - hívja a tartalmazott objektumok destruktorait
   - beállítja a virtuális függvények mutatóit
   - beállítja a virtuális alaposztály mutatóit
 * Hívja a közvetlen, nem virtuális alaposztályok destruktorait.
 * Öröklési lánc végén hívja a virtuális alaposztályok destruktorait.
 */


/// Virtuális alaposztály
struct VirtAlap {
    int* adat[3];   /// dump miatt 3 db int*
    VirtAlap() {
        for (int i = 0; i < 3; i++) adat[i] = reinterpret_cast<int*>(i+1);
        cout << "Oroklesi lanc vegen virt. alaposztaly konstruktor" << endl;
        g("VirtAlap()");
    }
    virtual ~VirtAlap() {
        cout << "Oroklesi lanc vegen alaposztaly destruktor" << endl;
        g("~VirtAlap()");
    }
    void g(const char *honnan) {
        cout << "  " << honnan << "-bol hivjuk f()-et:";
        f();
        dump();
    }
    virtual void f() {
        cout << " VirtAlap::f() vagyok" << endl;
    }
    /// p0 címtől l darab int*-ot ír ki
    void dump(void*p0, int l) {
        int **p = reinterpret_cast<int**>(p0);
        for (int i = 0; i < l; i++)
            cout << *p++ << " ";
        cout << endl;
    }

    void dump() {
#ifdef DUMP
        cout << "    Alaposztaly memoriakepe: ";
        dump(this, sizeof(VirtAlap)/sizeof(int*));
        int* p = *reinterpret_cast<int**>(this);
        cout << "    Virt.fv. tabla tartalma: ";
        dump(p, 2);
#endif
    }
};

/// Közvetlen nem virt alaposztály
struct KozvNemVirtAlap : virtual  VirtAlap {
    KozvNemVirtAlap()  {
        cout << "Kozvetlen nem virt. alaposztaly konstruktor" << endl;
        g("KozvNemVirtAlap()");
    }
    virtual ~KozvNemVirtAlap() {
        cout << "Kozvetlen nem virt. alaposztaly destruktor" << endl;
        g("~KozvNemVirtAlap()");
    }
    void f() {
        cout << " KozvNemVirtAlap::f() vagyok" << endl;
    }
};

/// Adattag
struct Adat {
    Adat()  { cout << "Tartalmazott obj. konstruktor" << endl; }
    ~Adat() { cout << "Tartalmazott obj. destruktor" << endl; }
};

/// Az OSZTALY
struct A : KozvNemVirtAlap {
    Adat adat;
    A()  {
        cout << "Konstruktor programozott torzse" << endl;
        g("A()");
    }
    ~A() {
        cout << "Destruktor programozott torzse" << endl;
        g("~A()");
    }
    void f() {
        cout << " A::f() vagyok" << endl;
    }
};

int main() {
    cout << "struct VirtAlap;\nstruct KozvNemVirtAlap : virtual VirtAlap;\n"
         << "struct A : KozvNemVirtAlap;\n"
         << "\nA* p = new A;\n\n";
    A* p = new A;
    cin.ignore(100, '\n');
    cout << "delete p;\n\n";
    delete p;
    return 0;
}

